/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.se.eip.patterns;

import java.net.URI;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;

import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.namespace.QName;
import javax.xml.transform.Source;
import javax.xml.transform.dom.DOMSource;
import javax.xml.xpath.XPathExpressionException;

import org.ow2.petals.component.framework.Constants.MEPConstants;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.util.SourceHelper;
import org.ow2.petals.component.framework.util.XMLUtil;
import org.ow2.petals.se.eip.ExchangeContext;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * Receives a stream of messages and identifies messages that are correlated
 * (XPATH expression). Once a complete set of messages has been received (XPATH
 * expression on special message to decide that the set is complete), the
 * aggregator collects information from each correlated message and publishes a
 * single, aggregated message to the output channel for further processing.
 * 
 * @author Adrien LOUIS - eBM WebSourcing
 * @author Christophe HAMERLING - eBM WebSourcing
 * @since 2.0
 * 
 */
public class Aggregator extends AbstractPattern {

    private Map<String, List<NormalizedMessage>> inMessagesMap;

    public static final String CORRELATION_ID = "aggregator-correlation";

    public static final String COMPLETE = "aggregator-complete";

    public static final String ROOT_ELEMENT = "aggregator-root-element";

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#process(org.ow2.petals
     * .component.framework.util.Exchange,
     * org.ow2.petals.se.eip.ExchangeContext)
     */
    public void process(Exchange exchange, ExchangeContext context) throws MessagingException,
            PEtALSCDKException {

        List<Consumes> consumes = context.getSUConsumes(exchange.getEndpoint());
        if (consumes.size() != 1) {
            throw new PEtALSCDKException(
                    "The Service Unit must define 1 Consumes section for the pattern "
                            + this.getPatternName());
        }

        // retrieve the document only once to avoid the Streaming issue
        Document inMessageContent = exchange.getInMessageContentAsDocument();

        String correlationID = retrieveCorrelationId(inMessageContent, exchange, context);
        if (correlationID == null) {
            throw new MessagingException("No correlation ID found in the message");
        }

        boolean complete = isComplete(inMessageContent, context);
        if (!this.inMessagesMap.containsKey(correlationID) && !complete) {
            this.inMessagesMap.put(correlationID, new ArrayList<NormalizedMessage>());
        }

        if (complete) {
            logger.info("Sequence is complete");
            Exchange response = sendAggregatedMessage(correlationID, context, consumes.get(0),
                    exchange.getOperation());
                PatternHelper.processSentExchange(response, exchange, context);
        } else {
            logger.info("Storing exchange '" + exchange.getExchangeId() + "' for correlation '"
                    + correlationID + "'");
            // reset the content of the in Message to avoid Streaming issue
            exchange.getInMessage().setContent(new DOMSource(inMessageContent));
            this.inMessagesMap.get(correlationID).add(exchange.getInMessage());
            logger.info("Sequence size is : " + this.inMessagesMap.get(correlationID).size());
        }
    }

    /**
     * Get the correlation ID from the message with an XPath expression or from
     * the exchange property.
     * 
     * @return
     */
    private String retrieveCorrelationId(Document inMessage, Exchange exchange,
            ExchangeContext context) {
        // retrieve the correlation from the message from an XPATH expression
        String result = null;
        String expression = context.getExtensions().get(CORRELATION_ID);
        if (expression != null) {
            try {
                result = PatternHelper.evaluateXPathExpression(expression, inMessage, context);
                logger.fine("Evaluate condition '" + expression + "', result '" + result + "'");
                if (Boolean.FALSE.toString().equals(result)) {
                    result = null;
                }

            } catch (XPathExpressionException e) {
                final String err = "Error in XPathExpression '" + expression + "', cause '"
                        + e.getCause().getMessage() + "'";
                logger.log(Level.SEVERE, err);
            }
        } else {
            // get the correlation from the Exchange properties
            result = (String) exchange.getProperty(CORRELATION_ID);
            if (result == null) {
                // get the correlation from the message properties
                result = (String) exchange.getInMessage().getProperty(CORRELATION_ID);
            }
        }
        return result;
    }

    /**
     * 
     * @param exchange
     * @return
     */
    private boolean isComplete(Document inMessage, ExchangeContext context) {
        boolean result = false;
        String condition = context.getExtensions().get(COMPLETE);
        try {
            result = PatternHelper.evaluateXPathCondition(condition, inMessage.getFirstChild(),
                    context);
            logger.fine("Evaluate condition: \"" + condition + "\", result: \"" + result + "\".");
        } catch (XPathExpressionException e) {
            final String err = "Error in XPathExpression \"" + condition + "\", cause: \""
                    + e.getCause().getMessage() + "\".";
            logger.log(Level.SEVERE, err);

        }
        return result;
    }

    /**
     * 
     * @param correlationId
     * @param context
     * @param consumes
     * @throws PEtALSCDKException
     */
    protected Exchange sendAggregatedMessage(String correlationId, ExchangeContext context,
            Consumes consumes, QName inOperation) throws MessagingException, PEtALSCDKException {

        List<NormalizedMessage> inMessageList = this.inMessagesMap.get(correlationId);
        if (inMessageList == null) {
            throw new MessagingException("No sequence found for correlation " + correlationId);
        }

        // aggregate the messages into a single one...
        Source content = null;
        try {
            content = buildContent(inMessageList, context);
        } catch (PEtALSCDKException e) {
            throw new MessagingException(e);
        }

        // send the aggregated message
        Exchange response = sendExchange(correlationId, consumes, content, context, inOperation);

        this.inMessagesMap.remove(correlationId);

        return response;
    }

    /**
     * Build the aggregate Source from list of messages.
     */
    private Source buildContent(List<NormalizedMessage> inMessagesList, ExchangeContext context)
            throws PEtALSCDKException {
        String root = getRootElement(context);

        Document aggregateDocument = XMLUtil.createDocument();
        Element rootElement = aggregateDocument.createElement(root);
        aggregateDocument.appendChild(rootElement);

        for (NormalizedMessage inMessage : inMessagesList) {
            Document inDocument = (Document) ((DOMSource) inMessage.getContent()).getNode();
            Node message = aggregateDocument.adoptNode(inDocument.getDocumentElement());
            rootElement.appendChild(message);
        }

        return SourceHelper.createSource(aggregateDocument);
    }

    /**
     * Get the root element name of the aggregated message
     * 
     * @param context
     * @return
     */
    private String getRootElement(ExchangeContext context) {
        String rootElement = context.getExtensions().get(ROOT_ELEMENT);

        if (rootElement == null) {
            // default value
            rootElement = getPatternName().toLowerCase();
        }

        return rootElement;
    }

    /**
     * Build the exchange to send to the target service.
     * 
     * @param correlationId
     * @param consumes
     * @param content
     * @param context
     * 
     * @return the result exchange
     * 
     * @throws MessagingException
     * @throws PEtALSCDKException
     */
    @SuppressWarnings("unchecked")
    private Exchange sendExchange(String correlationId, Consumes consumes, Source content,
            ExchangeContext context, QName inOperation) throws MessagingException,
            PEtALSCDKException {

        Exchange exchange = context.createConsumeExchange(consumes);
        NormalizedMessage aggregatedMessage = exchange.getInMessage();

        List<NormalizedMessage> inMessageList = this.inMessagesMap.get(correlationId);
        for (NormalizedMessage inMessage : inMessageList) {
            // setting properties
            for (String propertyName : (Set<String>) inMessage.getPropertyNames()) {
                aggregatedMessage.setProperty(propertyName, inMessage.getProperty(propertyName));
            }

            // setting attachments if any
            for (final String attachmentName : (Set<String>) inMessage.getAttachmentNames()) {
                aggregatedMessage.addAttachment(attachmentName, inMessage
                        .getAttachment(attachmentName));
            }
        }

        if (exchange.getOperation() == null) {
            exchange.setOperation(inOperation);
        }

        exchange.setInMessageContent(content);
        logSend(consumes, exchange);
        context.sendSync(exchange);

        return exchange;

    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#processMEPValidation(java
     * .net.URI)
     */
    protected boolean validateMEP(URI pattern) {
        return MEPConstants.IN_ONLY_PATTERN.equals(pattern)
                || MEPConstants.ROBUST_IN_ONLY_PATTERN.equals(pattern);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#init()
     */
    public void init() {
        this.inMessagesMap = new HashMap<String, List<NormalizedMessage>>();
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.AbstractPattern#getPatternName()
     */
    @Override
    protected String getPatternName() {
        return "Aggregator";
    }
}
