/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.se.eip.patterns;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;
import java.util.StringTokenizer;

import javax.activation.DataHandler;
import javax.jbi.JBIException;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.ow2.petals.component.framework.Constants.MEPConstants;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.jbidescriptor.generated.MEPType;
import org.ow2.petals.component.framework.util.InputStreamForker;
import org.ow2.petals.component.framework.util.XMLUtil;
import org.ow2.petals.se.eip.ExchangeContext;
import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * 
 * @author alouis
 * 
 */
public class PatternHelper {

    private static final XPath XPATH;
    static {
        XPATH = XPathFactory.newInstance().newXPath();
    }

    private static final String ROOT_ELEMENT = "root-element";

    private static final String FILTER = "filter";

    /**
     * Process the result of a piped exchange.<br>
     * - if the exchange contains an OUT response, return it and closed the
     * piped exchange<br>
     * - if the exchange contains a FAULT, set it on the original exchange and
     * return nothing, piped exchange is closed<br>
     * - if the exchange is ERROR, set it on the original exchange and return
     * nothing<br>
     * - if the exchange is DONE, return nothing<br>
     * 
     * @param sentExchange
     * @param originalExchange
     * @param context
     * @return the OUT message of the piped exchange, if any. When null, the
     *  original exchange has been updated
     * @throws JBIException
     */
    public static final NormalizedMessage processSentExchange(Exchange sentExchange,
            Exchange originalExchange, ExchangeContext context) throws MessagingException {
        NormalizedMessage receivedMessage = null;
        context.getLogger().fine(
                "Process exchange received from : srv="
                        + sentExchange.getEndpoint().getServiceName() + " / ep="
                        + sentExchange.getEndpoint().getEndpointName());
        if (sentExchange.isActiveStatus()) {
            if (sentExchange.getFault() != null) {
                context.getLogger().fine("a fault is returned, set it on the original Exchange.");
                // fault on piped exchange, set it on the original exchange
                // do not return a message
                Fault fault = originalExchange.createFault();
                fault.setContent(sentExchange.getFault().getContent());
                for (Object name : sentExchange.getFault().getPropertyNames()) {
                    fault.setProperty((String) name, sentExchange.getFault().getProperty(
                            (String) name));
                }
                for (Object name : sentExchange.getFault().getAttachmentNames()) {
                    fault.addAttachment((String) name, sentExchange.getFault().getAttachment(
                            (String) name));
                }
                originalExchange.setFault(fault);
            } else {
                // out message received, return it
                context.getLogger().info("an Out message is returned.");
                receivedMessage = sentExchange.getOutMessage();
                for (Object name : sentExchange.getOutMessage().getPropertyNames()) {
                    receivedMessage.setProperty((String) name, sentExchange.getOutMessage()
                            .getProperty((String) name));
                }
                for (Object name : sentExchange.getOutMessage().getAttachmentNames()) {
                    receivedMessage.addAttachment((String) name, sentExchange.getOutMessage()
                            .getAttachment((String) name));
                }
            }
            // close the exchange with the piped service
            sentExchange.setStatus(ExchangeStatus.DONE);
            context.send(sentExchange);
        } else if (sentExchange.isErrorStatus()) {
            // error on piped exchange, set it on the original exchange
            // do not return a message
            context.getLogger().fine(
                    "the exchange status is Error, set the original Exchange status to Error.");
            originalExchange.setError(sentExchange.getError());
        } else if (sentExchange.isDoneStatus()) {
            // piped service return nothing, exchange is closed
            // do not return a message
            context.getLogger().fine("the exchange status is Done, do nothing.");
        }

        return receivedMessage;
    }

    /**
     * Create an exchange for the given Consumes services. Set the IN message
     * with a <em>copy</em> of the given message
     * 
     * @param consumes
     * @param mep
     * @param originalMessage
     * @return
     * @throws JBIException
     * @throws PEtALSCDKException
     */
    public static final Exchange createExchange(Consumes consumes, MEPConstants mep,
            NormalizedMessage originalMessage, ExchangeContext context) throws MessagingException,
            PEtALSCDKException {
        Exchange tmpExchange = context.createConsumeExchange(consumes, mep);
        NormalizedMessage tmpMessage = tmpExchange.getInMessage();
        copy(originalMessage, tmpMessage);
        return tmpExchange;

    }

    /**
     * Copies a message's content, properties and attachments in another.
     * 
     * @param originalNm
     * @param copyNm
     * @throws MessagingException
     *      error while setting attachment
     * @throws PEtALSCDKException
     */
    @SuppressWarnings("unchecked")
    public static final void copy(final NormalizedMessage originalNm, final NormalizedMessage copyNm)
            throws MessagingException, PEtALSCDKException {

        // setting content
        copyNm.setContent(originalNm.getContent());

        // setting properties
        final Set<String> propertyNames = originalNm.getPropertyNames();
        for (final String propertyName : propertyNames) {
            copyNm.setProperty(propertyName, originalNm.getProperty(propertyName));
        }

        // setting attachments if any
        final Set<String> inAttachmentNameSet = originalNm.getAttachmentNames();
        if (inAttachmentNameSet != null) {
            for (final String attachmentName : inAttachmentNameSet) {
                final DataHandler attDh = originalNm.getAttachment(attachmentName);
                copyNm.addAttachment(attachmentName, attDh);
            }
        }

        // setting security context
        copyNm.setSecuritySubject(originalNm.getSecuritySubject());
    }

    /**
     * Retrieve the MEP from the extensions.
     * 
     * @param extensions
     * @return mep specified in extensions
     * @throws MessagingException
     *      MEP pattern specified in consumes extension isn't valid or if no MEP
     *      pattern specified in extensions.
     */
    public static final MEPConstants retrievePattern(MEPType patternShortcut)
            throws MessagingException {
        MEPConstants mep = null;

        if (patternShortcut != null) {
            if (MEPType.IN_ONLY.equals(patternShortcut)) {
                mep = MEPConstants.IN_ONLY_PATTERN;
            } else if (MEPType.ROBUST_IN_ONLY.equals(patternShortcut)) {
                mep = MEPConstants.ROBUST_IN_ONLY_PATTERN;
            } else if (MEPType.IN_OUT.equals(patternShortcut)) {
                mep = MEPConstants.IN_OUT_PATTERN;
            } else if (MEPType.IN_OPTIONAL_OUT.equals(patternShortcut)) {
                mep = MEPConstants.IN_OPTIONAL_OUT_PATTERN;
            }
        }

        return mep;
    }

    /**
     * 
     * @param condition
     * @param message
     * @param context
     * @return
     * @throws XPathExpressionException
     */
    public static final boolean evaluateXPathCondition(String expression, final Node message,
            ExchangeContext context) throws XPathExpressionException {
        String result = evaluateXPathExpression(expression, message, context);
        return result.equals("true");
    }

    /**
     * 
     * @param expression
     * @param message
     * @param context
     * @return
     * @throws XPathExpressionException
     */
    public static final String evaluateXPathExpression(String expression, final Node message,
            ExchangeContext context) throws XPathExpressionException {
        String result = "";
        final XPathExpression xpathExpression = XPATH.compile(expression);
        result = xpathExpression.evaluate(message);
        return result;
    }

    /**
     * Change the root element name of the input message
     * 
     * @param exchange
     * @param extensions
     */
    public static final void updateRootElement(Exchange exchange, ConfigurationExtensions extensions) {
        String rootElementName = extensions.get(ROOT_ELEMENT);
        if (rootElementName != null) {
            try {
                Document doc = exchange.getInMessageContentAsDocument();
                Element rootElement = doc.createElement(rootElementName);
                Element oldRootElement = doc.getDocumentElement();
                for (Node node : XMLUtil.getNodeChildren(oldRootElement)) {
                    rootElement.appendChild(node);
                }
                doc.replaceChild(rootElement, oldRootElement);
                exchange.setInMessageContent(doc);
            } catch (MessagingException e) {
                e.printStackTrace();
            } catch (DOMException e) {
                e.printStackTrace();
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    }

    /**
     * Filter all the children of the input message which match with the filter
     * element defined in the service unit.
     * 
     * @param exchange
     * @param extensions
     */
    public static final void filterElement(Exchange exchange, ConfigurationExtensions extensions) {
        String filter = extensions.get(FILTER);
        if (filter != null) {
            StringTokenizer tokenizer = new StringTokenizer(filter, ",");
            List<String> elements = new ArrayList<String>(tokenizer.countTokens());
            while (tokenizer.hasMoreTokens()) {
                elements.add(tokenizer.nextToken().trim());
            }

            try {
                Document doc = exchange.getInMessageContentAsDocument();
                Element oldRootElement = doc.getDocumentElement();
                Element newRootElement = doc.createElementNS(oldRootElement.getNamespaceURI(),
                        oldRootElement.getNodeName());
                newRootElement.setPrefix(oldRootElement.getPrefix());
                List<Node> children = XMLUtil.getNodeChildren(oldRootElement);

                for (Node node : children) {
                    String nodeName = node.getLocalName();
                    if (!elements.contains(nodeName)) {
                        newRootElement.appendChild(node);
                    }
                }
                doc.replaceChild(newRootElement, oldRootElement);
                exchange.setInMessageContent(doc);
            } catch (MessagingException e) {
                e.printStackTrace();
            } catch (DOMException e) {
                e.printStackTrace();
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    }

    /**
     * Handle the streaming issue.
     * 
     * @param streamSource
     * @throws MessagingException
     */
    public static final InputStreamForker handleStreamSource(NormalizedMessage message)
            throws MessagingException {
        InputStreamForker inMessageAsStream = null;
        if (message.getContent() instanceof StreamSource) {
            inMessageAsStream = new InputStreamForker(((StreamSource) message.getContent())
                    .getInputStream());
            Source inSource = new StreamSource(inMessageAsStream.getInputStreamOne());
            message.setContent(inSource);
        }

        return inMessageAsStream;
    }

    /**
     * Handle the streaming issue.
     * 
     * @param streamSource
     * @throws MessagingException
     */
    public static final InputStreamForker handleStreamSource(Fault fault) throws MessagingException {
        InputStreamForker faultMessageAsStream = null;
        if (fault.getContent() instanceof StreamSource) {
            faultMessageAsStream = new InputStreamForker(((StreamSource) fault.getContent())
                    .getInputStream());
            Source faultSource = new StreamSource(faultMessageAsStream.getInputStreamOne());
            fault.setContent(faultSource);
        }

        return faultMessageAsStream;
    }
}
